from typing import Union

import jax
from flax import linen as nn
from jax import numpy as jnp

from jaxformers.attention import MultiHeadAttention
from jaxformers.normalization import LayerNorm


class MLPBlock(nn.Module):
    in_features: int = 2048
    dropout_rate: float = 0.1

    @nn.compact
    def __call__(self, x: jax.Array, training: bool = False) -> jax.Array:
        assert x.ndim == 3, "x must be of shape [batch_size, seq_len, embed_dim]"
        assert x.dtype in [
            jnp.float16,
            jnp.float32,
            jnp.float64,
        ], "x must be of dtype float (float16, float32, float64)"
        out_features = x.shape[-1]

        x = nn.Dense(features=self.in_features)(x)
        x = nn.relu(x)
        x = nn.Dropout(rate=self.dropout_rate)(x, deterministic=not training)
        x = nn.Dense(features=out_features)(x)
        x = nn.Dropout(rate=self.dropout_rate)(x, deterministic=not training)
        return x


class EncoderBlock(nn.Module):
    embed_dim: int = 512
    num_heads: int = 8
    mlp_in_features: int = 2048
    dropout_rate: float = 0.1

    @nn.compact
    def __call__(self, x: jax.Array, training: bool = False) -> jax.Array:
        assert x.ndim == 3, "x must be of shape [batch_size, seq_len, embed_dim]"
        assert x.dtype in [
            jnp.float16,
            jnp.float32,
            jnp.float64,
        ], "x must be of dtype float (float16, float32, float64)"

        attention_out = MultiHeadAttention(
            num_heads=self.num_heads, embed_dim=self.embed_dim
        )(k=x, v=x, q=x)
        x += nn.Dropout(rate=self.dropout_rate)(
            attention_out, deterministic=not training
        )
        x = LayerNorm()(x)

        mlp_out = MLPBlock(in_features=self.mlp_in_features)(x)
        x += nn.Dropout(rate=self.dropout_rate)(mlp_out, deterministic=not training)
        return LayerNorm()(x)


class DecoderBlock(nn.Module):
    embed_dim: int = 512
    num_heads: int = 8
    mlp_in_features: int = 2048
    dropout_rate: float = 0.1

    @nn.compact
    def __call__(
        self,
        x: jax.Array,
        memory_k: jax.Array,
        memory_v: jax.Array,
        mask: Union[jax.Array, None] = None,
        training: bool = False,
    ) -> jax.Array:
        assert x.ndim == 3, "x must be of shape [batch_size, seq_len, embed_dim]"
        assert x.dtype in [
            jnp.float16,
            jnp.float32,
            jnp.float64,
        ], "x must be of dtype float (float16, float32, float64)"
        assert (
            memory_k.ndim == 3
        ), "memory_k must be of shape [batch_size, seq_len, embed_dim]"
        assert memory_k.dtype in [
            jnp.float16,
            jnp.float32,
            jnp.float64,
        ], "memory_k must be of dtype float (float16, float32, float64)"
        assert (
            memory_v.ndim == 3
        ), "memory_v must be of shape [batch_size, seq_len, embed_dim]"
        assert memory_v.dtype in [
            jnp.float16,
            jnp.float32,
            jnp.float64,
        ], "memory_v must be of dtype float (float16, float32, float64)"

        attention_out = MultiHeadAttention(
            num_heads=self.num_heads, embed_dim=self.embed_dim
        )(k=x, v=x, q=x, mask=mask)
        x += nn.Dropout(rate=self.dropout_rate)(
            attention_out, deterministic=not training
        )
        x = LayerNorm()(x)

        attention_out = MultiHeadAttention(
            num_heads=self.num_heads, embed_dim=self.embed_dim
        )(k=memory_k, v=memory_v, q=x, mask=mask)
        x += nn.Dropout(rate=self.dropout_rate)(
            attention_out, deterministic=not training
        )
        x = LayerNorm()(x)

        mlp_out = MLPBlock(in_features=self.mlp_in_features)(x)
        x += nn.Dropout(rate=self.dropout_rate)(mlp_out, deterministic=not training)
        return LayerNorm()(x)



from typing import Union

import flax.linen as nn
import jax
from jax import numpy as jnp

from jaxformers.blocks import DecoderBlock


class Decoder(nn.Module):
    num_layers: int = 12
    embed_dim: int = 512
    num_heads: int = 8
    mlp_in_features: int = 2048
    dropout_rate: float = 0.1

    @nn.compact
    def __call__(
        self,
        x: jax.Array,
        memory_k: jax.Array,
        memory_v: jax.Array,
        mask: Union[jax.Array, None] = None,
        training: bool = False,
    ) -> jax.Array:
        assert x.ndim == 3, "x must be of shape [batch_size, seq_len, embed_dim]"
        assert x.dtype in [
            jnp.float16,
            jnp.float32,
            jnp.float64,
        ], "x must be of dtype float (float16, float32, float64)"
        assert (
            memory_k.ndim == 3
        ), "memory_k must be of shape [batch_size, seq_len, embed_dim]"
        assert memory_k.dtype in [
            jnp.float16,
            jnp.float32,
            jnp.float64,
        ], "memory_k must be of dtype float (float16, float32, float64)"
        assert (
            memory_v.ndim == 3
        ), "memory_v must be of shape [batch_size, seq_len, embed_dim]"
        assert memory_v.dtype in [
            jnp.float16,
            jnp.float32,
            jnp.float64,
        ], "memory_v must be of dtype float (float16, float32, float64)"
        for _ in range(self.num_layers):
            x = DecoderBlock(
                embed_dim=self.embed_dim,
                num_heads=self.num_heads,
                mlp_in_features=self.mlp_in_features,
                dropout_rate=self.dropout_rate,
            )(x=x, memory_k=memory_k, memory_v=memory_v, mask=mask, training=training)
        return x



from typing import Tuple, Union

import jax
from flax import linen as nn
from jax import numpy as jnp


class SelfAttention(nn.Module):
    embed_dim: int = 512

    @nn.compact
    def __call__(
        self, k: jax.Array, v: jax.Array, q: jax.Array, return_attention: bool = False
    ) -> Union[jax.Array, Tuple[jax.Array, jax.Array]]:
        assert k.ndim == 3, "k must be of shape [batch_size, seq_len, embed_dim]"
        assert k.dtype in [
            jnp.float16,
            jnp.float32,
            jnp.float64,
        ], "k must be of dtype float (float16, float32, float64)"
        assert v.ndim == 3, "v must be of shape [batch_size, seq_len, embed_dim]"
        assert v.dtype in [
            jnp.float16,
            jnp.float32,
            jnp.float64,
        ], "v must be of dtype float (float16, float32, float64)"
        assert q.ndim == 3, "q must be of shape [batch_size, seq_len, embed_dim]"
        assert q.dtype in [
            jnp.float16,
            jnp.float32,
            jnp.float64,
        ], "q must be of dtype float (float16, float32, float64)"
        assert k.shape == v.shape, "k and v must have the same shape"
        assert k.shape == q.shape, "k and q must have the same shape"

        k_proj = nn.Dense(features=self.embed_dim)(k)
        v_proj = nn.Dense(features=self.embed_dim)(v)
        q_proj = nn.Dense(features=self.embed_dim)(q)

        energy = jnp.einsum("bqd,bkd->bqk", q_proj, k_proj)
        attention = nn.softmax(energy / jnp.sqrt(self.embed_dim), axis=-1)
        out = jnp.einsum("bqk,bkd->bqd", attention, v_proj)
        linear_out = nn.Dense(features=self.embed_dim)(out)
        if return_attention:
            return linear_out, attention
        return linear_out


class MultiHeadAttention(nn.Module):
    num_heads: int = 8
    embed_dim: int = 512
    head_dim: int = embed_dim // num_heads

    def setup(self) -> None:
        assert (self.head_dim * self.num_heads) == self.embed_dim, (
            "embed_dim must be equal to num_heads * head_dim, this means that embed_dim"
            " must be integer divisible by num_heads."
        )

    @nn.compact
    def __call__(
        self,
        k: jax.Array,
        v: jax.Array,
        q: jax.Array,
        mask: Union[jax.Array, None] = None,
        return_attention: bool = False,
    ) -> Union[jax.Array, Tuple[jax.Array, jax.Array]]:
        assert k.ndim == 3, "k must be of shape [batch_size, seq_len, embed_dim]"
        assert k.dtype in [
            jnp.float16,
            jnp.float32,
            jnp.float64,
        ], "k must be of dtype float (float16, float32, float64)"
        assert v.ndim == 3, "v must be of shape [batch_size, seq_len, embed_dim]"
        assert v.dtype in [
            jnp.float16,
            jnp.float32,
            jnp.float64,
        ], "v must be of dtype float (float16, float32, float64)"
        assert q.ndim == 3, "q must be of shape [batch_size, seq_len, embed_dim]"
        assert q.dtype in [
            jnp.float16,
            jnp.float32,
            jnp.float64,
        ], "q must be of dtype float (float16, float32, float64)"
        assert k.shape == v.shape, "k and v must have the same shape"
        assert k.shape == q.shape, "k and q must have the same shape"
        assert (
            mask is None or mask.ndim == 2
        ), "mask must be of shape [batch_size, seq_len]"

        batch_size, seq_len = q.shape[0], q.shape[1]

        k_proj = nn.Dense(features=self.embed_dim)(k)
        k_proj = k_proj.reshape((batch_size, self.num_heads, seq_len, self.head_dim))

        v_proj = nn.Dense(features=self.embed_dim)(v)
        v_proj = v_proj.reshape((batch_size, self.num_heads, seq_len, self.head_dim))

        q_proj = nn.Dense(features=self.embed_dim)(q)
        q_proj = q_proj.reshape((batch_size, self.num_heads, seq_len, self.head_dim))

        energy = jnp.einsum("bhqd,bhkd->bhqk", q_proj, k_proj)
        energy /= jnp.sqrt(self.embed_dim)
        if mask is not None:
            energy = jnp.where(mask == 0, -1e9, energy)
        attention = nn.softmax(energy, axis=-1)
        out = jnp.einsum("bhqk,bhkd->bhqd", attention, v_proj)
        out = out.reshape((batch_size, seq_len, self.embed_dim))
        linear_out = nn.Dense(features=self.embed_dim)(out)
        if return_attention:
            return linear_out, attention
        return linear_out



import jax
from flax import linen as nn
from jax import numpy as jnp


class PositionalEncoding(nn.Module):
    embed_dim: int = 512
    max_seq_len: int = 512

    def setup(self) -> None:
        self.positional_encoding = jnp.zeros((1, self.max_seq_len, self.embed_dim))
        x = jnp.arange(self.max_seq_len, dtype=jnp.float32).reshape(-1, 1) / jnp.power(
            10000, jnp.arange(0, self.embed_dim, 2, dtype=jnp.float32) / self.embed_dim
        )
        self.positional_encoding = self.positional_encoding.at[:, :, 0::2].set(
            jnp.sin(x)
        )
        self.positional_encoding = self.positional_encoding.at[:, :, 1::2].set(
            jnp.cos(x)
        )

    @nn.compact
    def __call__(self, x: jax.Array) -> jax.Array:
        assert x.ndim == 3, "x must be of shape [batch_size, seq_len, embed_dim]"
        assert x.dtype in [
            jnp.float16,
            jnp.float32,
            jnp.float64,
        ], "x must be of dtype float (float16, float32, float64)"
        return x + self.positional_encoding[:, : x.shape[1], :]



from typing import Union

import jax
from flax import linen as nn
from jax import numpy as jnp

from jaxformers.decoder import Decoder
from jaxformers.encoder import Encoder
from jaxformers.encoding import PositionalEncoding


class Transformer(nn.Module):
    embed_dim: int = 512
    max_seq_len: int = 512
    encoder_vocab_size: int = 256
    encoder_num_layers: int = 6
    encoder_num_heads: int = 8
    encoder_mlp_in_features: int = 2048
    encoder_dropout_rate: float = 0.1
    decoder_vocab_size: int = 256
    decoder_num_layers: int = 6
    decoder_num_heads: int = 8
    decoder_mlp_in_features: int = 2048
    decoder_dropout_rate: float = 0.1
    training: bool = False

    @nn.compact
    def __call__(
        self, x_in: jax.Array, x_out: jax.Array, mask_out: Union[jax.Array, None] = None
    ) -> jax.Array:
        assert x_in.ndim == 2, "x_in must be of shape [batch_size, seq_len]"
        assert x_in.dtype in [
            jnp.int8,
            jnp.int16,
            jnp.int32,
            jnp.int64,
        ], "x_in must be of dtype int (int8, int16, int32, int64)"
        assert x_out.ndim == 2, "x_out must be of shape [batch_size, seq_len]"
        assert x_in.dtype in [
            jnp.int8,
            jnp.int16,
            jnp.int32,
            jnp.int64,
        ], "x_out must be of dtype int (int8, int16, int32, int64)"

        x_in = nn.Embed(
            num_embeddings=self.encoder_vocab_size, features=self.embed_dim
        )(x_in)
        x_in = PositionalEncoding(
            embed_dim=self.embed_dim, max_seq_len=self.max_seq_len
        )(x_in)

        x_mem = Encoder(
            num_layers=self.encoder_num_layers,
            embed_dim=self.embed_dim,
            num_heads=self.encoder_num_heads,
            mlp_in_features=self.encoder_mlp_in_features,
            dropout_rate=self.encoder_dropout_rate,
        )(x=x_in, training=self.training)

        x_out = nn.Embed(
            num_embeddings=self.decoder_vocab_size, features=self.embed_dim
        )(x_out)
        x_out = PositionalEncoding(
            embed_dim=self.embed_dim, max_seq_len=self.max_seq_len
        )(x_out)

        out = Decoder(
            num_layers=self.decoder_num_layers,
            embed_dim=self.embed_dim,
            num_heads=self.decoder_num_heads,
            mlp_in_features=self.decoder_mlp_in_features,
            dropout_rate=self.decoder_dropout_rate,
        )(
            x=x_out,
            memory_k=x_mem,
            memory_v=x_mem,
            mask=mask_out,
            training=self.training,
        )

        out = nn.Dense(features=self.embed_dim)(out)
        return nn.softmax(out, axis=-1)



import flax.linen as nn
import jax
from jax import numpy as jnp

from jaxformers.blocks import EncoderBlock


class Encoder(nn.Module):
    num_layers: int = 12
    embed_dim: int = 512
    num_heads: int = 8
    mlp_in_features: int = 2048
    dropout_rate: float = 0.1

    @nn.compact
    def __call__(self, x: jax.Array, training: bool = False) -> jax.Array:
        assert x.ndim == 3, "x must be of shape [batch_size, seq_len, embed_dim]"
        assert x.dtype in [
            jnp.float16,
            jnp.float32,
            jnp.float64,
        ], "x must be of dtype float (float16, float32, float64)"
        for _ in range(self.num_layers):
            x = EncoderBlock(
                embed_dim=self.embed_dim,
                num_heads=self.num_heads,
                mlp_in_features=self.mlp_in_features,
                dropout_rate=self.dropout_rate,
            )(x=x, training=training)
        return x



import jax
from flax import linen as nn
from jax import numpy as jnp


class LayerNorm(nn.Module):
    eps: float = 1e-5

    @nn.compact
    def __call__(self, x: jax.Array) -> jax.Array:
        assert x.ndim > 1, "x must be of shape [batch_size, ...]"
        gamma = self.param("scale", nn.initializers.ones, x.shape[1:])
        beta = self.param("bias", nn.initializers.zeros, x.shape[1:])
        mean = jnp.mean(x, axis=-1, keepdims=True)
        std = jnp.std(x, axis=-1, keepdims=True)
        return jnp.asarray((gamma * (x - mean) / jnp.sqrt(std + self.eps)) + beta)
